"""Module implementaion of entrypoint API for Text2Text."""

import functools
import re
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from typing import Any, Dict, List, Optional

import mlflow
import pandas as pd
import requests as request_lib
from databricks.agents.evals import judges, metric  # pylint: disable=ungrouped-imports
from mlflow import create_experiment, get_experiment_by_name, get_trace, start_run
from mlflow.evaluation import Assessment
from mlflow.models import infer_signature
from tqdm import tqdm

from databricks.kie.t2t_custom_model import CustomModel, ModelWrapper
from databricks.kie.t2t_evaluation_criteria_generator import (add_evaluation_criteria_from_preference_data,
                                                              generate_evaluation_criteria)
from databricks.kie.t2t_instruction_analyzer import populate_instruction_info
from databricks.kie.t2t_metrics import create_classic_metrics
from databricks.kie.t2t_models.t2t_model_registrations import create_t2t_model_list
from databricks.kie.t2t_models.t2t_model_registry import T2TModelConfig
from databricks.kie.t2t_schema import SideBySideResult, T2TSystemParams, T2TSystemResult, Text2TextTaskSpec
from databricks.kie.t2t_utils import (INSTRUCTION_AND_INPUT_PROMPT_TEMPLATE, generate_exp_name_from_instruction,
                                      get_registered_model_name, validate_secret_exists)
from databricks.kie.task_spec import get_default_experiment_name_with_suffix
from databricks.model_training.api.utils import get_host_and_token_from_env
from databricks.model_training.mlflow_utils import get_mlflow_client
from databricks.model_training.version import __version__

_MODEL_CONCURRENCY = 3


@dataclass
class PredictionResult:
    """Container for model prediction results."""
    prediction: str
    trace: Optional[str] = None


class Text2TextRunner:
    """
    A class to manage the Text2Text model lifecycle, including training, evaluation, and deployment.

    Main API:
    - compile(): Optimizes prompt / model given provided hyperparameters.
    - update_system_param(): Updates the state of the runner with new instruction, evaluation criteria, or example data.
    - evaluate_models(): Evaluates all models in the runner.
    - visualize_models(): Visualizes model predictions for a subset of test data.
    - deploy_model(): Deploys a model to a production endpoint.
    """
    MAX_EVAL_DATA = 100
    MAX_EXAMPLES_FOR_VISUALIZATION = 5

    def __init__(self, task_spec: Text2TextTaskSpec):
        """
        Initializes the Text2TextRunner with the provided parameters.

        Args:
            task_spec: Text2TextTaskSpec object containing the instruction, example data, and evaluation criteria
        """
        self.model_map: Dict[str, ModelWrapper] = {}
        self.default_model: Optional[ModelWrapper] = None
        self.prev_task_spec: Optional[Text2TextTaskSpec] = None
        self.prev_params: Optional[T2TSystemParams] = None
        self.custom_metrics = []

        self.current_task_spec = task_spec
        self.current_params = task_spec.to_system_param()

        experiment_name = task_spec.experiment_name
        if not experiment_name:
            experiment_suffix = generate_exp_name_from_instruction(self.current_params.instruction)
            experiment_name = get_default_experiment_name_with_suffix(f"t2t_{experiment_suffix}")
            print(f"Experiment name not provided. Creating new experiment: {experiment_name}")

        self.experiment_id = self._get_exp_id(experiment_name)
        self.mlflow_client = get_mlflow_client()

    def compile(self) -> T2TSystemResult:
        # get common data shared between previous state and new state.
        common_data = []
        prev_res = []
        if self.prev_params:
            print("Generating response for previous state for side by side comparison...")
            # generate req/response for previous state. This will be used to compare with current state.
            common_data = self.prev_params.get_sxs_data(self.current_params)[0:self.MAX_EXAMPLES_FOR_VISUALIZATION]
            assert self.default_model is not None
            prev_res = self._visualize_model(self.default_model, data=common_data)

        print("Compiling new system...")
        # Construct evaluation criteria. Infer from examples if not provided.
        if not self.current_params.evaluation_criteria:
            print("Inferring evaluation criteria automatically...")
            inferred_eval_criteria = generate_evaluation_criteria(self.current_params.instruction,
                                                                  self.current_params.labelled_test_examples)
            self.current_params.evaluation_criteria = inferred_eval_criteria
            self.current_task_spec.evaluation_criteria = inferred_eval_criteria

        # If user provided labels or corrections, then add them to the evaluation criteria.
        if self.current_params.preference_data:
            new_eval_criteria = add_evaluation_criteria_from_preference_data(
                self.current_params.instruction,
                self.current_params.evaluation_criteria,
                self.current_params.preference_data,
            )
            self.current_params.evaluation_criteria = new_eval_criteria
            self.current_task_spec.evaluation_criteria = new_eval_criteria

        self._create_custom_metrics()
        populate_instruction_info(self.current_params)

        self._create_classic_metrics()
        self.create_models()

        # generate response for common data for current state -- to be used for SxS comparison.
        current_res = []
        if prev_res:
            assert common_data and self.default_model is not None
            print("Getting new system response for side by side comparison..")
            current_res = self._visualize_model(self.default_model, data=common_data)
        return self._get_model_result(prev_res, current_res)

    def _get_model_result(self, prev_result, current_result) -> T2TSystemResult:
        sxs_results = None
        if self.prev_params:
            sxs_results = []
            assert prev_result and current_result
            for prev, current in zip(prev_result, current_result):
                sxs_results.append(
                    SideBySideResult(request=prev['input'],
                                     previous_response=prev['model_output'],
                                     current_response=current['model_output']))

        assert self.default_model is not None

        print("Visualizing model results..")
        data = [{
            'request': datum[0],
            'response': datum[1]
        } for datum in self.current_params.labelled_test_examples[0:self.MAX_EXAMPLES_FOR_VISUALIZATION]]

        examples = self._visualize_model(self.default_model, data=data)

        print("Evaluating default model...")
        eval_result = self._evaluate_model(self.default_model, data=data)
        models = [model.run_name for model in self.model_map.values()]
        return T2TSystemResult(
            eval_result=eval_result,
            examples=examples,
            sxs_result=sxs_results,
            current_task_spec=self.current_task_spec,
            experiment_id=self.experiment_id,
            model_candidates=models,
            default_model=self.default_model.run_name,
            instruction_info=self.current_params.instruction_info,
        )

    def get_system_param(self) -> Dict[str, Any]:
        models = list(self.model_map.keys())

        return {
            "instruction": self.current_params.instruction,
            "eval_criteria": self.current_task_spec.evaluation_criteria,
            "example_data": self.current_params.labelled_test_examples[:self.MAX_EXAMPLES_FOR_VISUALIZATION],
            "experiment_id": self.experiment_id,
            "model_candidates": models,
            "instruction_info": self.current_params.instruction_info,
        }

    def update_system_param(self, new_spec: Text2TextTaskSpec):
        """Updates the state of the runner with new Text2Text spec."""
        if new_spec.experiment_name is not self.current_task_spec.experiment_name:
            raise ValueError("Experiment name cannot be updated.")
        new_system_param = new_spec.to_system_param()
        self.prev_params = self.current_params
        self.prev_task_spec = self.current_task_spec
        self.current_task_spec = new_spec
        self.current_params = new_system_param
        return self.compile()

    def evaluate_models(self):
        """Evaluates all models in the runner using the provided evaluation criteria."""
        result = {}
        if not self.current_task_spec.evaluation_criteria:
            raise ValueError("Evaluation criteria is not provided. Make sure to call `compile()` API.")

        for model_wrapper in self.model_map.values():
            model_eval = self._evaluate_model(model_wrapper)
            result[model_wrapper.run_name] = model_eval
        return result

    def visualize_models(self):
        """Runs a subset of test data through each model and returns the results."""
        result = {}
        for model_wrapper in self.model_map.values():
            result[model_wrapper.run_name] = self._visualize_model(model_wrapper)
        return pd.DataFrame.from_dict(result)

    def deploy_model(self, run_name: str, endpoint_name: str, secret_scope: str, secret_key: str):
        """Deploys a model to a production endpoint using the provided secret.

        Args:
            run_name: The run name of the model to deploy.
            endpoint_name: The name of the endpoint to deploy.
            secret_scope: The secret scope containing the secret.
            secret_key: The secret key to use for the deployment.
        """
        if not validate_secret_exists(secret_scope, secret_key):
            raise ValueError(f"Secret {secret_key} in scope {secret_scope} does not exist.")

        model_wrapper = None
        model_wrapper = self.model_map.get(run_name, None)
        if not model_wrapper:
            raise ValueError(
                f"invalid model descriptor.. input: `{run_name}` candidates: {list(self.model_map.keys())}")

        model_url = model_wrapper.get_model_path()
        production_model = mlflow.pyfunc.load_model(model_url)
        api_url, api_token = get_host_and_token_from_env()
        data = {
            "name": endpoint_name,
            "config": {
                "served_entities": [{
                    "entity_name": model_wrapper.registered_model_name,
                    "entity_version": '1',
                    "workload_size": "Small",
                    "scale_to_zero_enabled": False,
                    "environment_vars": {
                        "DATABRICKS_HOST": api_url,
                        "DATABRICKS_TOKEN": "{{" + f"secrets/{secret_scope}/{secret_key}" + "}}",
                    }
                }]
            }
        }

        headers = {"Context-Type": "text/json", "Authorization": f"Bearer {api_token}"}

        response = request_lib.post(url=f"{api_url}/api/2.0/serving-endpoints", json=data, headers=headers, timeout=60)
        print("Deployed endpoint.. Response status:", response.status_code)
        print("Deployed endpoint.. Response text:", response.text, "\n")
        return production_model

    def _log_model(self, model_wrapper: ModelWrapper):
        custom_model = model_wrapper.custom_model
        run_id = model_wrapper.run_id
        registered_model_name = get_registered_model_name(run_id)
        model_wrapper.registered_model_name = registered_model_name
        with start_run(experiment_id=self.experiment_id, run_id=run_id):
            mlflow.pyfunc.log_model(
                "model",
                python_model=custom_model,
                pip_requirements=["pandas", "mlflow", f"databricks-genai=={__version__}", "databricks-connect"],
                signature=custom_model.signature,
                registered_model_name=registered_model_name)

    def _get_exp_id(self, experiment_name: str):
        experiment = get_experiment_by_name(experiment_name)
        if not experiment:
            print(f"Experiment {experiment_name} not found. Creating new experiment...")
            return create_experiment(name=experiment_name)

        print(f"Experiment {experiment_name} found. Using existing experiment...")
        return experiment.experiment_id

    def _create_custom_metrics(self):
        """Create custom metrics for the model.

        Generates custom metrics for the model based on the self.current_task_spec.eval_criteria.
        Assumes that self.evaluation_criteria is already set.(i.e. compile() has been called)
        """

        def _wrapper_fn(request: str, response: str, guideline: str, metric_name: str):
            assessment = judges.guideline_adherence(
                request=request,
                response=response,
                guidelines=[guideline],
            )
            return Assessment(name=metric_name,
                              value=assessment.value,
                              rationale=assessment.rationale,
                              source=assessment.source)

        eval_criteria = self.current_task_spec.evaluation_criteria
        assert eval_criteria
        self.custom_metrics = []
        for criterion in eval_criteria:
            metric_name = re.sub(r'[^a-zA-Z0-9]', '_', criterion)
            custom_metric_fn = functools.partial(_wrapper_fn, guideline=criterion, metric_name=metric_name)
            custom_metric = metric(eval_fn=custom_metric_fn, name=metric_name)
            self.custom_metrics.append(custom_metric)

    def _create_classic_metrics(self):
        """Creates classic metrics for given the instruction type. 

        classic metrics are created based on the instruction type and includes 
        metrics such as fuzzy match & exact match.

        NOTE: Assumes self.current_params.instruction_info is already set.
        """
        assert self.current_params.instruction_info.instruction_type
        classic_metrics = create_classic_metrics(self.current_params.instruction_info.instruction_type)
        self.custom_metrics.extend(classic_metrics)

    def create_models(self):
        self.model_map.clear()
        model_list = create_t2t_model_list(self.current_params)
        for model_config in model_list:
            model = self._create_text2text_model(model_config)
            self.model_map[model.run_name] = model
            if model_config.is_default:
                self.default_model = model
            self._log_model(model)

    def _create_text2text_model(self, model_config: T2TModelConfig) -> ModelWrapper:
        with start_run(
                experiment_id=self.experiment_id,
                description=model_config.name,
        ) as run:
            signature = infer_signature(model_input=["input"], model_output=["output"])
            custom_model = CustomModel(
                model_config.model,
                signature,
                model_config.name,
            )
            run_id = run.info.run_id
            run_name = run.info.run_name
            return ModelWrapper(custom_model, run_id, run_name, registered_model_name=None)

    def _get_model_predictions(self,
                               model_wrapper: ModelWrapper,
                               requests: List[str],
                               with_traces: bool = False,
                               max_workers: int = _MODEL_CONCURRENCY) -> List[PredictionResult]:
        """Get predictions from model in parallel.

        Args:
            model_wrapper: ModelWrapper containing the model and metadata
            requests: List of requests to process
            with_traces: Whether to capture MLflow traces
            max_workers: Number of parallel workers

        Returns:
            List of PredictionResults containing predictions and optional traces
        """

        def predict_single(req: str) -> PredictionResult:
            if not with_traces:
                return PredictionResult(model_wrapper.custom_model.predict(None, req))

            root_span = self.mlflow_client.start_trace(name=model_wrapper.run_name)
            prediction = model_wrapper.custom_model.predict(None, req)
            self.mlflow_client.end_trace(root_span.request_id)
            return PredictionResult(prediction=prediction, trace=get_trace(root_span.request_id).to_json())

        desc = f"{'Evaluating' if with_traces else 'Visualizing'} model {model_wrapper.run_name}"
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            return list(tqdm(executor.map(predict_single, requests), total=len(requests), desc=desc))

    def _visualize_model(self, model_wrapper: ModelWrapper, data: Optional[List[Dict[str, str]]] = None):
        eval_data = data or [{
            'request': datum[0],
            'response': datum[1]
        } for datum in self.current_params.labelled_test_examples[0:self.MAX_EXAMPLES_FOR_VISUALIZATION]]
        requests = [datum['request'] for datum in eval_data]

        results = self._get_model_predictions(model_wrapper, requests)

        return [{
            "input": datum['request'],
            "model_output": result.prediction,
            **({
                "label": datum['response']
            } if 'response' in datum else {})
        } for datum, result in zip(eval_data, results)]

    def _evaluate_model(self,
                        model_wrapper: ModelWrapper,
                        data: Optional[List[Dict[str, str]]] = None) -> mlflow.models.EvaluationResult:
        eval_data = data or [{
            'request': datum[0],
            'response': datum[1]
        } for datum in self.current_params.labelled_test_examples[0:self.MAX_EVAL_DATA]]

        if not self.current_params.evaluation_criteria:
            raise ValueError("Evaluation criteria is not provided. Please invoke `compile()` API.")

        requests = [
            INSTRUCTION_AND_INPUT_PROMPT_TEMPLATE.format(instruction=self.current_params.instruction,
                                                         inp=datum['request']) for datum in eval_data
        ]

        with start_run(experiment_id=self.experiment_id, run_id=model_wrapper.run_id):
            results = self._get_model_predictions(model_wrapper, requests, with_traces=True)

            eval_df = pd.DataFrame({
                "request": requests,
                "expected_response": [datum['response'] for datum in eval_data],
                "response": [r.prediction for r in results],
                "trace": [r.trace for r in results]
            })

            return mlflow.evaluate(
                data=eval_df,
                model_type="databricks-agent",  # Enable Mosaic AI Agent Evaluation
                extra_metrics=self.custom_metrics,
                # Disable built-in judges.
                evaluator_config={'databricks-agent': {
                    "metrics": [],
                }})



"""Text2Text model that uses prompt tuning technique."""

import random
from typing import List, Optional, Tuple

from databricks.kie.inference_utils import get_llm_proxy_chat_completion_response
from databricks.kie.t2t_models.base_t2t_model import BaseT2TModel
from databricks.kie.t2t_schema import InstructionInfo, T2TSystemParams
from databricks.kie.t2t_utils import (create_chat_completions_messages_from_instruction,
                                      create_chat_completions_request, generate_cot_response)

MODEL_SELECTION_WITH_STRUCTURED_OUTPUT_SUPPORT = ["gpt-4o-2024-08-06-text2text", "gpt-4o-mini-2024-07-18-text2text"]


class T2TPromptTuningModel(BaseT2TModel):
    """A text-to-text model that uses prompt tuning with an underlying LLM.

    Args:
        instruction (str): The instruction guiding the model's behavior.
        model_id (str): The identifier of the underlying language model to use for generating predictions.
        few_shot_examples (Optional[List[Dict[str, str]]]): A list of few-shot examples for model signature inference.
        instruction_info (Optional[InstructionInfo]): The instruction info for the model.
        use_cot (bool): Whether to use chain-of-thought (CoT) for reasoning.
        temperature (Optional[float]): The temperature for sampling from the model.
    """
    INSTRUCTION_TEMPLATE = """
        INSTRUCTION: {optimized_instruction}
        OUTPUT FORMAT: {output_format}

        Given the following instruction and output format, provide response that follows the instruction and output format given the user input.
        User input will be provided below as part of the user message.  
    """
    MAX_ICL_EXAMPLES = 8

    def __init__(self,
                 instruction: str,
                 model_id: str,
                 few_shot_examples: Optional[List[Tuple[str, str]]] = None,
                 instruction_info: Optional[InstructionInfo] = None,
                 use_cot: bool = False,
                 temperature: Optional[float] = None):
        random.seed(0)
        self.instruction = instruction
        self.instruction_info = instruction_info
        self.model_id = model_id
        self.few_shot_examples = few_shot_examples or []
        self.temperature = temperature
        if len(self.few_shot_examples) > self.MAX_ICL_EXAMPLES:
            self.few_shot_examples = random.sample(self.few_shot_examples, self.MAX_ICL_EXAMPLES)

        self.use_cot = use_cot

    def __call__(self, model_input: str) -> str:
        instruction = self._get_instruction()
        messages = create_chat_completions_messages_from_instruction(instruction=instruction,
                                                                     inp=model_input,
                                                                     few_shot_examples=self.few_shot_examples)
        if self.use_cot:
            return generate_cot_response(self.model_id, messages, temperature=self.temperature)

        req = create_chat_completions_request(messages)
        return get_llm_proxy_chat_completion_response(self.model_id, req, temperature=self.temperature)

    def _get_instruction(self) -> str:
        if not self.instruction_info:
            return self.instruction

        optimized_instruction = self.instruction_info.optimized_instruction
        output_format = self.instruction_info.output_format
        return self.INSTRUCTION_TEMPLATE.format(optimized_instruction=optimized_instruction,
                                                output_format=output_format)

    @staticmethod
    def create_from_system_param(system_param: T2TSystemParams,
                                 use_cot: bool,
                                 model_id: str,
                                 temperature: Optional[float] = None) -> Optional['T2TPromptTuningModel']:
        if use_cot and model_id not in MODEL_SELECTION_WITH_STRUCTURED_OUTPUT_SUPPORT:
            return None

        instruction = (system_param.instruction_info.optimized_instruction
                       if system_param.instruction_info else system_param.instruction)
        few_shot_examples = system_param.labelled_training_examples
        return T2TPromptTuningModel(instruction=instruction,
                                    instruction_info=system_param.instruction_info,
                                    few_shot_examples=few_shot_examples,
                                    use_cot=use_cot,
                                    model_id=model_id,
                                    temperature=temperature)



"""Factory for KIE schemas"""
import hashlib
import json
from functools import reduce
from pathlib import Path
from types import UnionType
from typing import Any, Dict, List, Optional, Tuple, Type, TypeVar, Union, cast, get_args, get_origin

from pydantic import BaseModel, ConfigDict, Field, create_model
from pydantic.fields import FieldInfo

_valid_types = Union[Type, Type[Any], UnionType]

PYTHON_TO_YAML_TYPE_MAP: Dict[_valid_types, str] = {
    str: "str",
    int: "int",
    bool: "bool",
    float: "float",
    Any: "any",
}

YAML_TO_PYTHON_TYPE_MAP: Dict[str, _valid_types] = {v: k for k, v in PYTHON_TO_YAML_TYPE_MAP.items()}

JSON_TO_PYTHON_TYPE_MAP: Dict[str, _valid_types] = {
    "string": str,
    "integer": int,
    "bool": bool,
    "number": float,
    "any": Any,
}

CONFIG_DICT = ConfigDict(protected_namespaces=())


def create_custom_model(model_name: str, **field_definitions: Tuple[type, FieldInfo]) -> type[BaseModel]:
    """Create a Pydantic model with custom configuration."""
    return create_model(model_name, __config__=CONFIG_DICT, **field_definitions)  # type: ignore


class ModelFactory:
    """Factory class for creating Pydantic models from JSON examples."""

    @staticmethod
    def infer_type(value: Optional[Any], key: str = "DyanmicModel") -> Union[type, Any]:
        """Infer Python/Pydantic type from a JSON value."""
        if value is None:
            return Any

        if isinstance(value, bool):
            return bool
        elif isinstance(value, int):
            return int
        elif isinstance(value, float):
            return float
        elif isinstance(value, str):
            return str
        elif isinstance(value, dict):
            return ModelFactory.create_model_from_dict(value, key)
        elif isinstance(value, list):
            if not value:  # Empty list
                return List[Any]
            # Infer type from first element
            element_type = ModelFactory.infer_type(value[0], key)
            return List[element_type]
        else:
            # in theory, all of the above supports all the necessary JSON types
            raise NotImplementedError(f"Type inference for {value} is not implemented")

    @staticmethod
    def create_field_definitions(data: Dict[str, Any]) -> Dict[str, Tuple[type, FieldInfo]]:
        """Create field definitions for Pydantic model."""
        fields = {}
        for key, value in data.items():
            field_type = ModelFactory.infer_type(value, key)

            # Create field definition
            # Without knowing otherwise, we'll call every  field "Optional"
            fields[key] = (Optional[field_type], Field(description=""))

        return fields

    @staticmethod
    def create_model_from_dict(data: Dict[str, Any], model_name: str = "DynamicModel") -> type[BaseModel]:
        """Create a Pydantic model from a dictionary."""
        if data == {}:
            raise ValueError("Cannot create a model from an empty dictionary")

        field_definitions = ModelFactory.create_field_definitions(data)

        # Ignore typing: field_definitions keeps matching with other kwargs that are reserved.
        # Unsure how to clean that up
        return create_custom_model(model_name, **field_definitions)

    @classmethod
    def from_examples(cls, examples: List[Dict[str, Any]], model_name: str = "DynamicModel") -> type[BaseModel]:
        """Build and validate JSON schema from a dictionary of examples."""
        all_keys = reduce(lambda x, y: x.union(y), [set(e.keys()) for e in examples])
        merged_dict = ModelFactory._merge_and_validate_examples(examples, all_keys)

        return cls.create_model_from_dict(merged_dict, model_name)

    @staticmethod
    def to_file(model: Type[BaseModel], path: Union[Path, str]) -> None:
        """Save a Pydantic model class schema to file."""

        path = Path(path)
        path.parent.mkdir(parents=True, exist_ok=True)

        # Get full schema with definitions for nested models
        schema = model.model_json_schema()

        with path.open('w', encoding="utf-8") as f:
            json.dump({'_model_name': model.__name__, 'schema': schema}, f, indent=2)

    @classmethod
    def from_json_schema(cls, schema: Dict[str, Any], model_name: str = "DynamicModel") -> Type[BaseModel]:
        """
        Create a Pydantic model class from a JSON schema. Can optionally specify a model name.
        """

        def get_item_type(field_schema: Dict[str, Any]) -> _valid_types:
            """Get type from field schema, handling complex types like Optional[List[str]]."""

            # Handle anyOf (only allowed for Optional types)
            if 'anyOf' in field_schema:
                types = field_schema['anyOf']
                # Check if this is an Optional type (one null and one non-null)
                if len(types) == 2 and any(t.get('type') == 'null' for t in types):
                    # Get the non-null type - e.g. {'type': 'string'}
                    non_null_type = next(t for t in types if t.get('type') != 'null')
                    found_type = get_item_type(non_null_type)
                    return Optional[found_type]
                else:
                    # Raise an error for other Union types
                    raise TypeError("Union types are only allowed for Optional fields."
                                    f"Found schema: {field_schema}")

            # Handle direct type definitions
            if 'type' in field_schema:
                found_type = field_schema['type']

                # Handle arrays
                if found_type == 'array' and 'items' in field_schema:
                    if '$ref' in field_schema['items']:
                        ref_name = field_schema['items']['$ref'].split('/')[-1]
                        ref_schema = schema.get('$defs', {}).get(ref_name, {})
                        item_type = create_model_from_schema(ref_schema, ref_name)
                    else:
                        item_type = get_item_type(field_schema['items'])
                    return List[item_type]

                # Handle basic types
                return JSON_TO_PYTHON_TYPE_MAP.get(found_type, Any)

            return Any

        def create_model_from_schema(schema_data: Dict[str, Any], model_name: str) -> Type[BaseModel]:
            fields = {}

            for field_name, field_schema in schema_data.get('properties', {}).items():
                # Handle references to nested models
                if '$ref' in field_schema:
                    ref_name = field_schema['$ref'].split('/')[-1]
                    ref_schema = schema.get('$defs', {}).get(ref_name, {})
                    field_type = create_model_from_schema(ref_schema, ref_name)
                else:
                    field_type = get_item_type(field_schema)

                # Create field with description and title if present
                field_info = {}
                if 'description' in field_schema:
                    field_info['description'] = field_schema['description']
                if 'title' in field_schema:
                    field_info['title'] = field_schema['title']

                # Setup the type, Field tuple
                # Note all fields are considered "required", but some are nullable.
                fields[field_name] = (field_type, Field(**field_info))

            return create_custom_model(model_name, **fields)

        return create_model_from_schema(schema, model_name)

    @classmethod
    def from_file(cls, path: Union[Path, str]) -> Type[BaseModel]:
        """
        Load a Pydantic model class from a schema file.
        Expects the file to be in the format saved by `to_file`, that is,
        it should have a JSON schema and a model name.
        """

        path = Path(path)
        with path.open('r', encoding="utf-8") as f:
            data = json.load(f)

        return cls.from_json_schema(data['schema'], data['_model_name'])

    @staticmethod
    def _merge_and_validate_examples(examples: List[Dict[str, Any]], all_keys: set[str]) -> Dict[str, Any]:
        """Merge and validate examples.
        
        The output of this is expected to be a valid dictionary such that all potential conflicts
        are solved and that Nones are removed.
        """

        def _handle_lists(list_examples: List[List[Any]]) -> List[Any]:
            flattened_list = list(filter(lambda z: z is not None, reduce(lambda x, y: x + y, list_examples)))

            if len(flattened_list) == 0:
                raise ValueError(f"Field {key} is missing examples or all values are None")

            inner_types = {type(e) for e in flattened_list}
            if len(inner_types) > 1:
                raise ValueError(f"Field {key} has conflicting types in its lists: {inner_types}")

            if isinstance(flattened_list[0], dict):
                list_dict_keys = set(flattened_list[0].keys())
                if any(set(e.keys()) != list_dict_keys for e in flattened_list):
                    raise ValueError(f"Field {key} has conflicting keys across its list of dictionaries")

                return [ModelFactory._merge_and_validate_examples(flattened_list, list_dict_keys)]
            elif inner_types == {list}:
                return [_handle_lists(flattened_list)]
            return [flattened_list[0]]

        merged_dict = {}
        for key in all_keys:
            key_examples = [e[key] for e in examples if key in e and e[key] is not None]

            if len(key_examples) == 0:
                raise ValueError(f"Miss non-empty examples for the field {key}")

            types = {type(e) for e in key_examples}
            if len(types) > 1:
                raise ValueError(f"Field {key} has conflicting types: {types}")
            field_type = types.pop()

            if field_type == dict:
                dict_keys = set(key_examples[0].keys())
                for key_example in key_examples:
                    if set(key_example.keys()) != dict_keys:
                        raise ValueError(f"Field {key} has conflicting keys across its dictionaries")

                merged_dict[key] = ModelFactory._merge_and_validate_examples(key_examples, set(key_examples[0].keys()))
            elif field_type == list:
                merged_dict[key] = _handle_lists(key_examples)
            else:
                merged_dict[key] = key_examples[0]
        return merged_dict


T = TypeVar('T', bound=BaseModel)


class ModelManager:
    """Manager for editing fields on a Pydantic model"""

    def __init__(self,
                 model: Type[T],
                 parent: Optional['ModelManager'] = None,
                 field_name: Optional[str] = None) -> None:
        self.model = model
        self.parent = parent
        self.field_name = field_name

        if (self.parent is None) is not (self.field_name is None):
            raise ValueError("Must specify both `parent` and `field_name` if one of them is specified")

    def __getitem__(self, field_name: str) -> 'ModelManager':
        if field_name not in self.model.model_fields:
            raise KeyError(f"Field {field_name} not found")

        field = self.model.model_fields[field_name]
        if not (isinstance(field.annotation, type) and issubclass(field.annotation, BaseModel)):
            raise TypeError(f"Field {field_name} is not a nested model")

        return ModelManager(field.annotation, self, field_name)

    @staticmethod
    def _copy_field(f: FieldInfo, **updates) -> Tuple[type, FieldInfo]:
        annotation = updates.pop("annotation", f.annotation)
        copied = {
            "description": f.description,
            "title": f.title,
            "default": f.default,
            "json_schema_extra": f.json_schema_extra,
        }
        copied.update(**updates)
        return (annotation, Field(**copied))

    def _update_model_with_fields(self, **fields: Tuple[type, FieldInfo]) -> Type[BaseModel]:
        new_model = create_custom_model(f"{self.model.__name__}", **fields)  # type: ignore
        self.model = cast(Type[T], new_model)
        return self.model

    def _update_parents(self) -> Type[BaseModel]:
        if not self.parent:
            return self.model

        current = self
        while current.parent is not None:
            assert current.field_name is not None
            current.parent.edit_types(**{current.field_name: current.model})
            current = current.parent
        return current.model

    def add_field(self, field_name: str, field_type: type, field_description: str = "") -> type[BaseModel]:

        # Add existing fields
        fields = {name: self._copy_field(field) for name, field in self.model.model_fields.items()}

        # Add new field
        fields[field_name] = (field_type, Field(description=field_description))

        self._update_model_with_fields(**fields)

        return self._update_parents()

    def edit_types(self, **type_updates: type) -> Type[BaseModel]:

        fields = {
            name: self._copy_field(field, annotation=type_updates.get(name, field.annotation))
            for name, field in self.model.model_fields.items()
        }
        self._update_model_with_fields(**fields)

        return self._update_parents()

    def edit_descriptions(self, **updates) -> Type[BaseModel]:

        fields = {
            name: self._copy_field(field, description=updates.get(name, field.description))
            for name, field in self.model.model_fields.items()
        }
        self._update_model_with_fields(**fields)

        return self._update_parents()


def model_to_dict(model: Type[BaseModel]) -> Dict[str, Any]:
    """Convert a Pydantic model to a dictionary describing its structure.

    Args:
        model: A Pydantic model class
        
    Returns:
        Dict with field names as keys and dicts containing:
            - type: String for basic types, dict for nested models, list for arrays 
            - description: Field description string
            
    Example:
        class User(BaseModel):
            name: str = Field(description="User name")
            tags: List[str]
            
        # Returns:
        {
            'name': {'type': 'str', 'description': 'User name'},
            'tags': {'type': [{'type': 'str'}], 'description': ''}
        }
    """
    output = {}
    for key, field in model.model_fields.items():
        type_ = field.annotation
        value: Dict[str, Any] = {"description": field.description or "", "required": True}
        if get_origin(type_) is Union:  # This is an optional field
            type_ = [t for t in get_args(type_) if t is not None][0]
            value["required"] = False

        if get_origin(type_) is list:
            args = get_args(type_)
            if isinstance(args[0], type) and issubclass(args[0], BaseModel):
                # This is an array of models
                type_value = [model_to_dict(args[0])]
            else:
                type_value = [{"type": PYTHON_TO_YAML_TYPE_MAP.get(args[0], "any")}]
        elif isinstance(type_, type) and issubclass(type_, BaseModel):
            # This is a nested model
            type_value = model_to_dict(type_)
        else:
            type_value = PYTHON_TO_YAML_TYPE_MAP.get(type_, "any")
        value['type'] = type_value
        output[key] = value
    return output


def dict_to_model(model_dict: Dict[str, Any]) -> Type[BaseModel]:
    """Create a Pydantic model from a dictionary describing its structure.

    Args:
        model_dict: Dict with field names as keys and field info dicts containing:
            - type: String for basic types, dict for nested models, list for arrays
            - description: Optional field description string
            
    Returns:
        A generated Pydantic model class with the specified structure
        
    Example:
        schema = {
            'name': {'type': 'str', 'description': 'User name'},
            'tags': {'type': [{'type': 'str'}]}
        }
        UserModel = dict_to_model(schema)
    """
    fields = {}
    for key, value in model_dict.items():
        description = value.get('description', '')
        required = value.get('required', True)

        requested_type = value.get('type')
        if not requested_type:
            raise ValueError(f"Field {key} is missing a type, please specify one in your schema.")

        type_ = Any
        if isinstance(requested_type, list):
            item = requested_type[0]
            if 'type' in item:
                requested_inner_type = item['type']
                if requested_inner_type not in YAML_TO_PYTHON_TYPE_MAP:
                    raise ValueError(f"Array field {key} has an unknown type: {requested_inner_type}. " +
                                     "Please update your schema to provide a valid type.")
                inner_type = YAML_TO_PYTHON_TYPE_MAP.get(requested_inner_type, Any)
                if inner_type == Any:
                    raise ValueError(f"Array field {key} has a type of any, which is not supported. " +
                                     "Please update your schema to provide a specific type.")

                type_ = List[inner_type]
            else:
                item_type = dict_to_model(item)
                type_ = List[item_type]
        elif isinstance(requested_type, dict):
            # Nested model
            type_ = dict_to_model(requested_type)
        elif requested_type in YAML_TO_PYTHON_TYPE_MAP:
            # Basic type
            type_ = YAML_TO_PYTHON_TYPE_MAP.get(requested_type)
        else:
            raise ValueError(
                f"Field {key} has an unknown type: {requested_type}. Please update your schema to provide a valid type."
            )

        if type_ == Any:
            raise ValueError(f"Field {key} has a type of any, which is not supported. " +
                             "Please update your schema to provide a specific type.")

        if not required:
            type_ = Optional[type_]
        fields[key] = (type_, Field(description=description))

    return create_custom_model("Model", **fields)


def get_schema_hash(response_format: Type[BaseModel]) -> str:
    schema = json.dumps(response_format.model_json_schema())
    return hashlib.md5(schema.encode()).hexdigest()[:8]
